#!/usr/bin/env python
# Copyright (C) 2012  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import code
import readline
import rlcompleter
import __builtin__
from lmi.lmi_client_base import LmiReturnValue
from lmi.lmi_client_shell import LmiConnection
from lmi.lmi_client_shell import _LmiNamespace
from lmi.lmi_client_shell import _LmiClass
from lmi.lmi_client_shell import _LmiInstance

def _lmi_displayhook(o):
    if o == None:
        return
    __builtin__._ = None
    if isinstance(o, LmiReturnValue):
        sys.stdout.write("%s: %s (%d)" % (o.hostname, "ok" if o.rval == 0 else "fail", o.rval))
        if o.errorstr:
            sys.stdout.write(": %s" % o.errorstr)
        sys.stdout.write("\n")
    else:
        print o
    __builtin__._ = o

sys.displayhook = _lmi_displayhook

class _LmiShellOptions(object):
    def __init__(self, argv):
        self.m_argv = argv
        self.m_script_name = ""
        self.m_script_argv = []
        self.m_help = False
        for a in self.m_argv[1:]:
            if not self.m_script_name and (a == "-h" or a == "--help"):
                self.m_help = True
                return
            elif not self.m_script_name and not a.startswith("-"):
                self.m_script_name = a
            else:
                self.m_script_argv.append(a)

    @property
    def help(self):
        return self.m_help

    @property
    def interactive(self):
        return not self.m_script_name

    @property
    def scriptName(self):
        return self.m_script_name

    @property
    def scriptArgs(self):
        return [self.m_script_name] + self.m_script_argv

    def printUsage(self):
        sys.stdout.write("Usage: %s [script] [script-options]\n" % os.path.basename(self.m_argv[0]))
        sys.stdout.write("\nOptions:\n  -h, --help            print this message\n")

class _LmiCompleter(rlcompleter.Completer):
    def __init__(self, namespace = None):
        rlcompleter.Completer.__init__(self, namespace)
        self.m_last_complete = []

    def __fetchClasses(self, client, expr):
        namespace = expr.name.split("/")[-1]
        if self.m_cache[client][namespace].classes:
            return
        result = {}
        classes = expr.classes()
        for c in classes:
            result[c] = _LmiCompleterClassCache()
        self.m_cache[client][namespace].classes = result

    def __completeObjectMethods(self, obj, text):
        result = filter(lambda m: not m.startswith("_") \
            and not m.startswith("m_") \
            and callable(getattr(obj, m)) \
            and m.lower().startswith(text.lower()), dir(obj))
        return result

    def __completeObjectProperties(self, obj, text):
        result = filter(lambda m: not m.startswith("_") \
            and not m.startswith("m_") \
            and not callable(getattr(obj, m)) \
            and m.lower().startswith(text.lower()), dir(obj))
        return result

    def _callable_postfix(self, val, word):
        if hasattr(val, '__call__') \
            and not isinstance(val, (_LmiNamespace, _LmiClass, _LmiInstance)):
            word = word + "("
        return word

    def complete(self, text, state):
        if not text:
            return ("\t", None)[state]
        if state > 0:
            return self.m_last_complete[state]
        self.m_last_complete = []
        members = text.split(".")
        if members[0] in self.namespace:
            cmd = ".".join(members[0:-1])
            to_complete = members[-1]
            expr = eval(cmd, self.namespace)
            if isinstance(expr, LmiConnection):
                namespaces = expr.namespaces()
                for n in namespaces:
                    if n.lower().startswith(to_complete.lower()):
                        self.m_last_complete.append(cmd + "." + n)
                return self.m_last_complete[state]
            elif isinstance(expr, _LmiNamespace):
                for c in expr.classes():
                    if c.lower().startswith(to_complete.lower()):
                        self.m_last_complete.append(cmd + "." + c)
                methods = self.__completeObjectMethods(expr, to_complete)
                self.m_last_complete.extend(cmd + "." + m + "(" for m in methods)
                properties = self.__completeObjectProperties(expr, to_complete)
                self.m_last_complete.extend(cmd + "." + p for p in properties)
                return self.m_last_complete[state]
            elif isinstance(expr, _LmiClass):
                methods = self.__completeObjectMethods(expr, to_complete)
                self.m_last_complete= [cmd + "." + m + "(" for m in methods]
                properties = self.__completeObjectProperties(expr, to_complete)
                self.m_last_complete.extend(cmd + "." + p for p in properties)
                return self.m_last_complete[state]
            elif isinstance(expr, _LmiInstance):
                for m in expr.methods():
                    if m.lower().startswith(to_complete.lower()):
                        self.m_last_complete.append(cmd + "." + m + "(")
                for p in expr.properties():
                    if p.lower().startswith(to_complete.lower()):
                        self.m_last_complete.append(cmd + "." + p)
                methods = self.__completeObjectMethods(expr, to_complete)
                self.m_last_complete.extend(cmd + "." + m + "(" for m in methods)
                properties = self.__completeObjectProperties(expr, to_complete)
                self.m_last_complete.extend(cmd + "." + p for p in properties)
                return self.m_last_complete[state]
        return rlcompleter.Completer.complete(self, text, state)

class LmiInteractiveShell(code.InteractiveConsole):
    def __init__(self, prompt, more_prompt, locals = None):
        sys.ps1 = prompt
        sys.ps2 = more_prompt
        if locals == None:
            locals = {}
        locals["LmiConnection"] = LmiConnection
        locals["connect"] = _connect_interactive
        locals["clear_history"] = self.clearHistory
        locals["use_display_sugar"] = _use_display_sugar
        code.InteractiveConsole.__init__(self, locals)
        self.m_completer = _LmiCompleter(self.locals)
        self.m_history_path = os.path.expanduser("~/.lmi_shell_history")
        readline.set_completer(self.m_completer.complete)
        readline.parse_and_bind('tab: complete')
    
    def interact(self):
        self.loadHistory()
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = self.raw_input(prompt)
                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, unicode):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\n")
                self.resetbuffer()
                more = 0
        self.saveHistory()
    
    def loadHistory(self):
        if not os.path.exists(self.m_history_path):
            return
        readline.read_history_file(self.m_history_path)

    def saveHistory(self):
        readline.write_history_file(self.m_history_path)

    def clearHistory(self):
        readline.clear_history()

def _lmi_interact(prompt = "> ", more_prompt = "... "):
    console = LmiInteractiveShell(prompt, more_prompt)
    console.interact()

def _connect_interactive(hostname, username = "", password = ""):
    return _connect(hostname, username, password, True)

def _connect_noninteractive(hostname, username = "", password = ""):
    return _connect(hostname, username, password, False)

def _connect(hostname, username = "", password = "", interactive = False):
    if not username:
        while True:
            try:
                username = raw_input("username: ")
                if username:
                    break
            except EOFError, e:
                sys.stdout.write("\n")
                continue
            except KeyboardInterrupt, e:
                sys.stdout.write("\n")
                return None
        readline.remove_history_item(readline.get_current_history_length() - 1)
    if not password:
        try:
            os.system("stty -echo")
            password = raw_input("password: ")
            os.system("stty echo")
        except EOFError, e:
            password = ""
        except KeyboardInterrupt, e:
            os.system("stty echo")
            sys.stdout.write("\n")
            return None
        sys.stdout.write("\n")
        readline.remove_history_item(readline.get_current_history_length() - 1)
    return LmiConnection(hostname, username, password, interactive)

def _use_display_sugar(use = True):
    sys.displayhook = _lmi_displayhook if use else sys.__displayhook__

if __name__ == "__main__":
    options = _LmiShellOptions(sys.argv)
    if options.help:
        options.printUsage()
        sys.exit(0)

    if options.interactive:
        _lmi_interact()
    else:
        locals = {
            "LmiConnection" : LmiConnection,
            "connect" : _connect_noninteractive
        }
        sys.argv = options.scriptArgs
        execfile(options.scriptName, locals)
    sys.exit(0)
