# Copyright (C) 2012  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import pywbem
import collections

_RValue = collections.namedtuple('RValue', ["rval", "rparams", "errorstr"])

class LmiReturnValue(_RValue):
    @property
    def hostname(self):
        return self.m_hostname

    @hostname.setter
    def hostname(self, hostname):
        self.m_hostname = hostname

    def __new__(self, **kwargs):
        self.m_hostname = "none"
        if "hostname" in kwargs:
            self.m_hostname = kwargs["hostname"]
            kwargs.pop("hostname")
        return _RValue.__new__(self, **kwargs)

class LmiBaseClient(object):
    def __init__(self, hostname, username = "", password = ""):
        self.m_hostname = hostname
        self.m_username = username
        self.m_password = password
        self.m_cliconn = pywbem.WBEMConnection("https://" + self.m_hostname,
            (self.m_username, self.m_password))

    def _getInstanceNames(self, klass, **kwargs):
        try:
            inst_name_list = self.m_cliconn.EnumerateInstanceNames(klass)
        except pywbem.cim_operations.CIMError, e:
            return (None, e.args[1])
        except pywbem.cim_http.AuthError, e:
            return (None, e.args[0])
        return (inst_name_list, "")

    def _getInstances(self, klass, filter_key = "", **kwargs):
        (inst_name_list, errorstr) = self._getInstanceNames(klass, **kwargs)
        if not inst_name_list:
            return (None, errorstr)
        inst_names = inst_name_list
        if len(filter_key):
            key = kwargs["Key"] if "Key" in kwargs else "Name"
            inst_names = filter(lambda n: n[key] == filter_key, inst_name_list)
        inst_list = []
        for i in inst_names:
            inst_list.append(self.m_cliconn.GetInstance(i, LocalOnly = False))
        if not len(inst_list):
            return (None, "Not found" if not "ErrorMsg" in kwargs else kwargs["ErrorMsg"])
        return (inst_list, "")

    def _deleteInstance(self, obj):
        try:
            self.m_cliconn.DeleteInstance(obj.path)
        except pywbem.cim_operations.CIMError, e:
            return (False, e.args[1])
        return (True, "")

    def _getClassNames(self, namespace = None, **kwargs):
        try:
            class_name_list = self.m_cliconn.EnumerateClassNames(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            return (None, e.args[1])
        except pywbem.cim_http.AuthError, e:
            return (None, e.args[0])
        return (class_name_list, "")

    def _getClasses(self, namespace = None, **kwargs):
        try:
            class_list = self.m_cliconn.EnumerateClasses(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            return (None, e.args[1])
        except pywbem.cim_http.AuthError, e:
            return (None, e.args[0])
        return (class_list, "")

    def _getClass(self, class_name, namespace = None, **kwargs):
        try:
            klass = self.m_cliconn.GetClass(class_name, namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            return (None, e.args[1])
        except pywbem.cim_http.AuthError, e:
            return (None, e.args[0])
        return (klass, "")
    
    def _callMethodRaw(self, instance, method, **params):
        try:
            (rval, rparams) = self.m_cliconn.InvokeMethod(method, instance.path, **params)
        except pywbem.cim_operations.CIMError, e:
            return (-1, {}, e.args[1] + ": '" + method + "'")
        return (rval, rparams, "")

    def _callMethod(self, instance, method, **params):
        (rval, rparams, errorstr) = self._callMethodRaw(instance, method, **params)
        return (rval == 0, rparams, errorstr)

    def _getAssociators(self, instance, **params):
        return self.m_cliconn.Associators(instance.path, **params)
