# Copyright (C) 2012  Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import new
import sys
import pywbem
import textwrap
import tempfile
import subprocess
from xml.dom.minicompat import NodeList
from lmi_client_base import LmiBaseClient
from lmi_client_base import LmiReturnValue

class _LmiErrorHandler(object):
    def __init__(self, errormsg):
        super(self.__class__, self).__init__()
        if errormsg:
            sys.stderr.write("Error: %s\n" % errormsg)

    def __getattr__(self, name):
        return _LmiErrorHandler("")

    def __call__(self, *args, **kwargs):
        return _LmiErrorHandler("")

    def __iter__(self):
        return _LmiErrorHandler("")

    def __getitem__(self, key):
        return _LmiErrorHandler("")

    def __str__(self):
        return ""

class _LmiTextFormatter(object):
    def __init__(self, text):
        self.m_text = text

    def format(self, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno(), **kwargs):
        if indent > width:
            return # NOTE: this is wrong!
        wrapper = textwrap.TextWrapper()
        wrapper.width = width - indent
        wrapper.subsequent_indent = " " * sub_indent
        for l in wrapper.wrap(self.m_text):
            os.write(fd, "%s%s\n" % (" " * indent, l))
        if "separator" in kwargs and kwargs["separator"]:
            os.write(fd, "\n")

class _LmiXmlFormatter(object):
    def __init__(self, xml):
        self.m_xml = xml
        self.m_width = int(os.popen('stty size', 'r').read().split()[1])
        if not self.m_width:
            self.m_width = 80

    def __formatClassPropertyContent(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        children = node.childNodes
        children_len = len(children)
        for (i, n) in enumerate(children):
            if isinstance(n, pywbem.cim_xml.QUALIFIER):
                self.__formatQualifier(n, indent, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__formatClassProperty(n, indent, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__formatClassPropertyArray(n, indent, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.METHOD):
                self.__formatMethod(n, indent, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.VALUE):
                # NOTE: we skip default value for the class
                pass

    def __formatQualifier(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        is_array = isinstance(node.firstChild, pywbem.cim_xml.VALUE_ARRAY)
        val = "[qualifier] %s%s %s" % (node.getAttribute("TYPE"), " []" if is_array else "",
            node.getAttribute("NAME"))
        values = node.childNodes
        if values and isinstance(values[0], pywbem.cim_xml.VALUE_ARRAY):
            val += ": { " + ", ".join(["'" + v.firstChild.nodeValue + "'" for v in values[0].childNodes]) + " }\n"
        else:
            val += ": '%s'\n" % values[0].firstChild.nodeValue
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, fd, separator = True)

    def __formatClassProperty(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        val = "[property] %s %s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, fd, separator = True)
        self.__formatClassPropertyContent(node, indent + 4, sub_indent, width, fd)

    def __formatClassPropertyArray(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        val = "[property array] %s [] %s" % (node.getAttribute("TYPE"), node.getAttribute("TYPE"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, fd, separator = True)
        self.__formatClassPropertyContent(node, indent + 4, sub_indent, width, fd)

    def __formatClassPropertyReference(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        val = "[property ref] %s %s" % (node.getAttribute("REFERENCECLASS"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, fd, separator = True)
        self.__formatClassPropertyContent(node, indent + 4, sub_indent, width, fd)
    
    def __formatInstanceProperty(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        val = node.firstChild
        prop = "[property] %s %s%s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"),
            " = '%s'" % val.firstChild.nodeValue if val else "")
        _LmiTextFormatter(prop).format(indent, sub_indent + 4, width, fd, separator = True)

    def __formatInstancePropertyArray(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        prop = "[property array] %s [] %s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"))
        val = node.firstChild
        if val:
            prop += " = { " + ", ".join(["'" + v.firstChild.nodeValue + "'" for v in val.childNodes]) + " }"
        _LmiTextFormatter(prop).format(indent, sub_indent + 4, width, fd, separator = True)

    def __formatInstancePropertyReference(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        val = "[property ref] %s %s" % (node.getAttribute("REFERENCECLASS"), node.getAttribute("NAME"))
        _LmiTextFormatter(val).format(indent, sub_indent + 4, width, fd, separator = True)

    def __formatMethod(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        qualifiers = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.QUALIFIER)]
        parameters = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER)]
        parameters_arr = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_ARRAY)]
        parameters_ref = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_REFERENCE)]
        parameters_ref_arr = [x for x in node.childNodes if isinstance(x, pywbem.cim_xml.PARAMETER_REFARRAY)]
        has_args = parameters != [] or parameters_arr != [] or parameters_ref != [] or parameters_ref_arr != []
        val = "[method] %s %s%s" % (node.getAttribute("TYPE"), node.getAttribute("NAME"), "(...)" if has_args else "()")
        _LmiTextFormatter(val).format(indent, sub_indent, width, fd, separator = True)
        for q in qualifiers:
            self.__formatQualifier(q, indent + 4, sub_indent + 4, width, fd)
        for p in parameters:
            val = "[param] %s %s" % (p.getAttribute("TYPE"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, fd, separator = True)
        for p in parameters_arr:
            val = "[param array] %s (ref) %s" % (p.getAttribute("TYPE"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, fd, separator = True)
        for p in parameters_ref:
            val = "[param ref] %s (ref) %s" % (p.getAttribute("REFERENCECLASS"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, fd, separator = True)
        for p in parameters_ref_arr:
            val = "[param ref array] %s (ref) %s" % (p.getAttribute("REFERENCECLASS"), p.getAttribute("NAME"))
            _LmiTextFormatter(val).format(indent + 4, sub_indent + 4, width, fd, separator = True)

    def __formatClass(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        _LmiTextFormatter("Class: %s" % node.getAttribute("NAME")).format(indent, sub_indent, width, fd)
        _LmiTextFormatter("SuperClass: %s" % node.getAttribute("SUPERCLASS")).format(indent + 4, sub_indent, width, fd)
        for n in node.childNodes:
            if isinstance(n, pywbem.cim_xml.QUALIFIER):
                self.__formatQualifier(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__formatClassProperty(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__formatClassPropertyArray(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_REFERENCE):
                self.__formatClassPropertyReference(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.METHOD):
                self.__formatMethod(n, indent + 4, sub_indent, width, fd)

    def __formatInstance(self, node, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        instance = node.firstChild.nextSibling
        _LmiTextFormatter("Instance of %s" % instance.getAttribute("CLASSNAME")).format(indent, sub_indent, width, fd)
        for n in instance.childNodes:
            if isinstance(n, pywbem.cim_xml.PROPERTY):
                self.__formatInstanceProperty(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_ARRAY):
                self.__formatInstancePropertyArray(n, indent + 4, sub_indent, width, fd)
            elif isinstance(n, pywbem.cim_xml.PROPERTY_REFERENCE):
                self.__formatInstancePropertyReference(n, indent + 4, sub_indent, width, fd)
   
    @staticmethod
    def __which(program):
        is_exe = lambda fpath: os.path.isfile(fpath) and os.access(fpath, os.X_OK)
        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None

    def __getPager(self):
        if "PAGER" in os.environ:
            return os.environ["PAGER"]
        for p in ["less", "more"]:
            path = self.__which(p)
            if path:
                return path

    def format(self, indent = 0, sub_indent = 0, width = 80, fd = sys.stdout.fileno()):
        if isinstance(self.m_xml, pywbem.cim_xml.CLASS):
            self.__formatClass(self.m_xml, indent, sub_indent, width, fd)
        elif isinstance(self.m_xml, pywbem.cim_xml.VALUE_NAMEDINSTANCE):
            self.__formatInstance(self.m_xml, indent, sub_indent, width, fd)

    def fancyFormat(self, interactive):
        if interactive:
            tmpfile = tempfile.mkstemp()
            self.format(0, 0, self.m_width, tmpfile[0])
            os.close(tmpfile[0])
            subprocess.call([self.__getPager(), tmpfile[1]])
            os.remove(tmpfile[1])
        else:
            self.format(0, 0, self.m_width)

# Base class for classes, that are used for tab-completion.
# Used due to deprecated methods since python 2.2.
class BaseObject(object):
    # deprecated
    def __methods__(self):
        return []

    # deprecated
    def __members__(self):
        return []
    
    def __call__(self):
        pass

class _LmiNamespace(BaseObject):
    def __init__(self, client, name):
        super(self.__class__, self).__init__()
        self.m_client = client
        self.m_name = name

    def __getattr__(self, name):
        if name == "name":
            return self.m_name
        (klass, errorstr) = self.m_client.getClass(name, self.m_name)
        if not klass:
            return _LmiErrorHandler("Error: %s" % errorstr)
        return _LmiClass(self.m_client, self, klass)

    def classes(self, filter_key = "", exact_match = False):
        (class_name_list, errorstr) = self.m_client.getClassNames(self.m_name)
        if not class_name_list:
            if errorstr:
                sys.stderr.write("Error: %s\n" % errorstr)
            else:
                sys.stderr.write("No classes present in namespace '%s'\n" % self.m_name)
            return
        if filter_key:
            if not exact_match:
                filter_lambda = lambda n: n.lower().find(filter_key.lower()) >= 0
            else:
                filter_lambda = lambda n: n.lower() == filter_key.lower()
            class_name_list = filter(filter_lambda, class_name_list)
        return class_name_list

    def printClasses(self, filter_key = "", exact_match = False):
        for c in self.classes(filter_key, exact_match):
            sys.stdout.write("%s\n" % c)

class _LmiClass(BaseObject):
    def __init__(self, client, namespace, klass):
        super(self.__class__, self).__init__()
        self.m_client = client
        self.m_namespace = namespace
        self.m_klass = klass

    def doc(self):
        _LmiXmlFormatter(self.m_klass.tocimxml()).fancyFormat(self.m_client.interactive)
       
    def instances(self, filter_key = "", **kwargs):
        instance_list = self.m_client.getInstances(self.m_klass.classname, filter_key, **kwargs)
        if not instance_list:
            return _LmiErrorHandler("'%s' instance does not exist" % filter_key)
        result = []
        for i in instance_list:
            result.append(_LmiInstance(self.m_client,  self, i))
        return result

    def firstInstance(self, filter_key = "", **kwargs):
        instance_list = self.instances(filter_key, **kwargs)
        if not instance_list:
            return _LmiErrorHandler("'%s' instance does not exist" % filter_key)
        return instance_list[0]

    def properties(self):
        return self.m_klass.properties.keys()

    def printProperties(self):
        for (name, prop) in self.m_klass.properties.iteritems():
            sys.stdout.write("%s\n" % name)

    def methods(self):
        return self.m_klass.methods.keys()

    def printMethods(self):
        for (name, method) in self.m_klass.methods.iteritems():
            sys.stdout.write("%s\n" % name)

    @property
    def classname(self):
        return self.m_klass.classname

    @property
    def namespace(self):
        return self.m_namespace.name

class _LmiInstance(BaseObject):
    class _LmiInstanceMethod(BaseObject):
        def __init__(self, client, instance, method):
            super(self.__class__, self).__init__()
            self.m_client = client
            self.m_instance = instance
            self.m_method = method

        def __call__(self, *args, **kwargs):
            for (param, value) in kwargs.iteritems():
                if param in self.m_method.parameters:
                    t = self.m_method.parameters[param].type.lower()
                    cast = {
                        "sint8"  : lambda x: pywbem.Sint8(x),
                        "uint8"  : lambda x: pywbem.Uint8(x),
                        "sint16" : lambda x: pywbem.Sint16(x),
                        "uint16" : lambda x: pywbem.Uint16(x),
                        "sint32" : lambda x: pywbem.Sint32(x),
                        "uint32" : lambda x: pywbem.Uint32(x),
                        "sint64" : lambda x: pywbem.Sint64(x),
                        "uint64" : lambda x: pywbem.Uint64(x)
                    }
                    if t in cast:
                        cast_func = cast[t]
                        kwargs[param] = cast_func(value)
            return self.m_client.callMethodRaw(
                self.m_instance, self.m_method.name, **kwargs)

    def __init__(self, client, klass, inst):
        super(self.__class__, self).__init__()
        self.m_client = client
        self.m_klass = klass
        self.m_inst = inst

    def __str__(self):
        return str(self.m_inst)

    def __getattr__(self, name):
        methods = self.__methods()
        if methods and name in methods:
            return _LmiInstance._LmiInstanceMethod(self.m_client, self, self.__method(name))
        elif not name in self.m_inst.properties.keys():
            return _LmiErrorHandler("'%s' is not valid property of the class '%s'" % (name, self.m_klass.classname))
        return self.m_inst.properties[name].value

    def __getitem__(self, key):
        return self.__getattr__(key)
    
    def __methods(self):
        (klass, errorstr) = self.m_client.getClass(self.m_inst.classname, self.m_klass.namespace)
        if not klass:
            return []
        return klass.methods

    def __method(self, name):
        methods = self.__methods()
        if not methods:
            return None
        try:
            return methods[name]
        except KeyError, e:
            return None

    @property
    def classname(self):
        return self.m_klass.classname

    @property
    def namespace(self):
        return self.m_klass.namespace

    @property
    def path(self):
        return self.m_inst.path
    
    def doc(self):
        _LmiXmlFormatter(self.m_inst.tocimxml()).fancyFormat(self.m_client.interactive)

    def associators(self):
        result = []
        associators_list = self.m_client.getAssociators(self.m_inst)
        for i in associators_list:
            (klass, errorstr) = self.m_client.getClass(i.classname, self.namespace)
            if not klass:
                return []
            lmi_namespace = self.m_klass.m_namespace
            lmi_class = _LmiClass(self.m_client, lmi_namespace, klass)
            lmi_assoc_inst = _LmiInstance(self.m_client, lmi_class, i)
            result.append(lmi_assoc_inst)
        return result

    def properties(self):
        return self.m_klass.properties()

    def printProperties(self):
        self.m_klass.printProperties()

    def methods(self):
        return self.m_klass.methods()

    def printMethods(self):
        self.m_klass.printMethods()

class _LmiShellClient(LmiBaseClient):
    def __init__(self, hostname, username = "", password = "", interactive = False):
        super(self.__class__, self).__init__(hostname, username, password)
        self.m_interactive = interactive

    def getClassNames(self, namespace = None):
        return self._getClassNames(namespace, DeepInheritance = True)

    def getInstances(self, class_name, filter_key = "", **kwargs):
        (instance_list, _) = self._getInstances(class_name, filter_key, **kwargs)
        if not instance_list:
            return None
        return instance_list

    def getClass(self, class_name, namespace = None):
        return self._getClass(class_name, namespace)
    
    def getAssociators(self, instance):
        return self._getAssociators(instance)

    def callMethodRaw(self, instance, method, **params):
        (rv, rp, err) = self._callMethodRaw(instance, method, **params)
        return LmiReturnValue(hostname = self.m_hostname, rval = rv, rparams = rp, errorstr = err)

    @property
    def interactive(self):
        return self.m_interactive

class LmiConnection(object):
    def __init__(self, hostname, username = "", password = "", interactive = False):
        self.m_shell_client = _LmiShellClient(hostname, username, password, interactive)

    def namespaces(self):
        return ["root", "cimv2", "interop"]

    def printNamespaces(self):
        sys.stdout.write("root\n")
        sys.stdout.write("cimv2\n")
        sys.stdout.write("interop\n")

    @property
    def root(self):
        return _LmiNamespace(self.m_shell_client, "root")

    @property
    def cimv2(self):
        return _LmiNamespace(self.m_shell_client, "root/cimv2")

    @property
    def interop(self):
        return _LmiNamespace(self.m_shell_client, "root/interop")
